// Timetable Generator API Include File
// 
//
// Copyright (C) 2012, Botond Ballo.
//
// This library is free software. You can redistribute and/or modify 
// it under the terms of the Timetable Generator API License, 
// version 2.0 or later (see accompanying file LICENSE.txt).

// Note: while this file is reasonably well-commented, it is highly
// recommended that users of this API read the HTML documentation
// accompanying this library (built with "make docs") which provides
// an overview of the library as well as more detailed information
// about each API type and function.

#ifndef LIBTG_INCLUDE_TG_API_HPP_INCLUDE_GUARD
#define LIBTG_INCLUDE_TG_API_HPP_INCLUDE_GUARD

#include <memory>      // std::unique_ptr
#include <string>      // std::string
#include <ostream>     // std::ostream
#include <vector>      // std::vector
#include <utility>     // std::move, std::pair
#include <map>         // std::map
#include <functional>  // std::function
#include <istream>     // std::istream
#include <ostream>     // std::ostream
#include <chrono>      // std::chrono::seconds

#include <cstdint>  // uint64_t

namespace tg
{
    namespace detail
    {
    
        using std::unique_ptr;
        using std::string;
        using std::ostream;
        using std::vector;
        using std::move;
        using std::pair;
        using std::map;
        using std::function;
        using std::istream;
        using std::ostream;
        using std::chrono::seconds;
        
        // Public types
        
        // enum class term_type
        // This enumeration is used to classify timetable pages.
        // A timetable contains one page for each term (or two pages
        // for each term if the timetable includes alternate-week labs).
        enum class term_type {fall, spring};
        
        // enum class session_type
        // This enumeration is used to classify courses as running in
        // the fall term (F), in the spring term (S), or both terms (Y).
        enum class session_type {F = 0, S, Y};
        
        // enum class timetable_sort_criterion
        // This enumeration is used to represent a criterion by which
        // timetables are to be sorted. The actual values of this 
        // enumeration are an implementation detail of the library. 
        enum class timetable_sort_criterion : size_t;
        
        // enum class course_load_state
        // This enumeration is used to divide a user's interaction with a
        // course load into distinct actions (selecting a study session,
        // selecting courses, selecting configurations, and generating
        // timetables) and keep track of what phase of this process a
        // course load is in.
        // The enumeration values are given numeric values so that
        // relational comparisons can be done on them (e.g. <=).
        enum class course_load_state 
        {
            empty = 0, 
            session_selected = 1, 
            courses_selected = 2, 
            configs_selected = 3, 
            timetables_generated = 4
        };
        
        // class timetable_sort_criteria
        // Represents a list of criteria (each of type timetable_sort_criterion)
        // by which timetables are to be sorted. Timetables are sorted first
        // by the first criterion; then, within a group of timetables that rank
        // as equal by the first criterion, timetables are sorted by the second
        // criterion; and so on.
        typedef vector<timetable_sort_criterion> timetable_sort_criteria;
        
        // struct ClassTime
        // Represents a time when a meeting starts or ends.
        // The current granularity is 30 minutes.
        struct ClassTime
        {
            // overloaded operators
            void operator++();  // increment by 30 minutes
            void operator--();  // decrement by 30 minutes

            // text output
            string to_str() const;

            // data members
            unsigned hour;  // can range from 9 to 21
            bool     half;  // add 30 minutes to "hour" if set
        };

        // operator-(ClassTime a, ClassTime b)
        // Find the number of "unit times" (30-minute periods) between a and b.
        size_t operator-(const ClassTime& a, const ClassTime& b);
        
        // struct Block
        // Represents one block of time during which a class runs.
        struct Block
        {
            unsigned  day;         // day of week (0 = Monday, 4 = Friday)
            ClassTime start_time;  // start time of block
            size_t    length;      // length of block in unit blocks (currently 1 unit block = 30 min)
        };

        // struct CourseBlocks
        // Information about a single meeting section in a timetable page.
        // Includes information about the meeting section (section code, instructor)
        // and the course to which it belongs (activity code), and all the
        // blocks of time during which the meeting section meets.
        struct CourseBlocks
        {
            string        activity_code;     // activity code for these blocks
            string        section_code;      // section code for these blocks
            string        instructor;        // instructor for these blocks
            string        instructor_short;  // shortened version of 'instructor' 
            vector<Block> blocks;            // the time blocks themselves
        };

        // struct TimetablePage
        // Represents one page of a timetable. Timetables contain one page
        // for each term (or two pages for each term if there are
        // alternate-week labs in the timetable).
        struct TimetablePage
        {
            // convenience accessors
            string name() const;  // name of the page (e.g. "Fall Term" or "Spring Term, Week 1")

            // data members
            term_type            term;       // fall or spring term
            bool                 alternate;  // is this page specific to one of the alternating weeks?
            unsigned             week;       // if so, which week?
            vector<CourseBlocks> courses;    // the blocks of time in this timetable, categorized by meeting section
                                             // i.e. one element of this vector stores the blocks for one meeting section
        };

        // struct Timetable
        // Represents one timetable.
        struct Timetable
        {
            // accessors
            
            // Get the i'th page of this timetable.
            const TimetablePage& operator[](size_t i) const { return pages[i]; }
            
            
            // static methods
            
            // Get the number of pages in this timetable.
            size_t num_pages() const { return pages.size(); }
            
            // Number of pages in a timetable based on whether or not is has alternate-week sections.
            // This will be 4 if the timetable has alternate-week sections, or 2 if it does not.
            // Note, if *any* timetable in a set of generated timetables has alternate-week section,
            // *each* timetable in the set will have 4 pages).
            static size_t num_pages(bool has_alternate_weeks);


            // data members
            vector<TimetablePage> pages;
        };
        
        // struct Offering
        // Represents a single offering a course.
        struct Offering
        {
            string       activity_code;
            session_type session_code;
            
            // Get a string combining the activity code and the session code.
            string activity_session_code() const;
            
            bool operator==(const Offering& other) const
            {
                return activity_code == other.activity_code && session_code == other.session_code;
            }
        };
        
        // struct Offerings
        // Represents all the offerings of a course in a session of study.
        struct Offerings
        {
            string               key;  // to pass to SelectCourses()
            string               activity_code;
            string               title;
            vector<session_type> sessions_offered;
            
            // Get a string combining the activity code and title.
            string get_activity_code_and_title() const;
            
            // Get the activity code portion of the combined (activity code + title)
            // returned by get_activity_code_and_title().
            static string get_activity_code_part(const string& activity_code_and_title);
            
            // Get the department code of a course from its activity code.
            // This can be passed either the activity code itself, or the string
            // returned by get_activity_code_and_title().
            static string get_department_code(const string& activity_code);
        };
        
        // struct CourseList
        // Represents all the offerings of all courses in a session of study
        struct CourseList
        {
            map<string, vector<Offerings>> departments;  // department name -> courses in department
        };
        
        // struct ActivityConfiguration
        // Represents a single activity configuration (choice of offering for each course in a course load).
        struct ActivityConfiguration
        {
            vector<Offering> offerings;                // offerings in configuration
            bool             has_alternate_timeslots;  // does this configuration need separate pages for alternate weeks?
            
            bool operator==(const ActivityConfiguration& other) const
            {
                return has_alternate_timeslots == other.has_alternate_timeslots && offerings == other.offerings;
            }
        };
        
        // struct StudySession
        // Represents information about a session of study
        struct StudySession
        {
            string name;          // name of session (e.g. "2010 Summer")
            string faculty_list;  // faculties available in this study session
            bool   is_default;    // is this the default study session?
        };
        
        // struct GenerationStatistics
        // A class that stores various statistics about the timetable generation process
        // which may be displayed to the user during generation.
        struct GenerationStatistics
        {
            // constructor
            GenerationStatistics() : timetables_generated(0), average_generation_speed(0) {}
            
            // data members
            size_t   timetables_generated;      // number of timetables generated so far
            seconds  elapsed_time;              // elapsed time since beginning of generation processed
            unsigned average_generation_speed;  // in timetables per second
        };
        
        // struct FileSaveState
        // Stores the name of the file where a course load is saved
        // (if any), and whether the course load has changed since
        // the last time it was saved to this file.
        struct FileSaveStatus
        {
            string filename;
            bool   up_to_date;
        };
        
        // struct TimetableDataSource, TimetableDataSink
        // TimetableDataSource and TimetableDataSink are abstractions of a "data store"
        // from which timetable data can be read and to which it can be written, respectively.
        // The interface for reading is to get an istream corresponding to a given "key".
        // The interface for writing is to get an ostream corresponding to a given key.
        // The most obvious implementation is a filesystem, with the streams being file streams
        // and the keys being filenames.
        struct TimetableDataSource
        {
            virtual ~TimetableDataSource() {}
            virtual istream& get_reader(const string& key) = 0; 
        };
        struct TimetableDataSink
        {
            virtual ~TimetableDataSink() {}
            virtual ostream& get_writer(const string& key) = 0;
        };
        
        // struct ProgressIndicator
        // Used to report the progress of time consuming API calls.
        // 'callback' is a function the API call will call from time to time,
        // with an integer argument 'arg' in the range [0, 'granularity'], to
        // indicate that 'arg'/'granularity' of the process is complete.
        // The callback will be called no more than 'granularity' + 1 times
        // in total.
        struct ProgressIndicator
        {
            function<void(int)> callback;
            size_t              granularity;
        };

        // Utility functions
        // These are part of the API
        // They are completely stateless, requiring neither per-session nor global state
        
        // Get the earliest possible start time for a class
        ClassTime GetEarliestClassTime();
        
        // Get the total number of course blocks from morning to evening
        size_t GetTotalTimeBlocks();
        
        // Get a list of possible criteria to be used when sorting timetables, along with their descriptions.
        // Note: the list of criteria depends on whether or not the activity configuration whose timetables
        // are being sorted has alternate timeslots. 
        map<timetable_sort_criterion, string> GetTimetableSortCriteria(bool has_alternate_timeslots);
        
        // Check connectivity to the university websites from which timetable data is downloaded.
        bool CheckUniversityWebsiteConnectivity();
        
        // Forward declarations
        class course_load_impl;
        class user_session_impl;
        class timetable_manager_impl;
        class TimetableManager;
        
        // class CourseLoad
        // This is the interface to API functions operating on a course load.
        class CourseLoad
        {
        public:
            // Make class non-copyable but movable.
            // This is so that UserSession::NewCourseLoad() and UserSession::OpenCourseLoad()
            // can return a CourseLoad by value and thus avoid having to manage the lifetime
            // of the course load itself.
            CourseLoad(CourseLoad&&);
            CourseLoad& operator=(CourseLoad&&);
            
            // Destructor
            // Public because UserSession::NewCourseLoad() and UserSession::OpenCourseLoad()
            // return a CourseLoad by value.
            ~CourseLoad();
            
            // API functions
            
            // Determine whether the course load has been modified since
            // the last time it was modified (false), or if it is up to date (true).
            bool IsUpToDate() const;
            
            // Get the name of the file the course load has last been saved to.
            // Returns the empty string if the course load has never been saved.
            string GetFileName() const;
            
            // Save the course load to the given file.
            void Save(const string& filename);
            
            // Get the current course load state.
            course_load_state GetState() const;
            
            // Get the currently selected study session.
            string GetCurrentStudySession() const;

            // Select a study session.
            void SelectStudySession(size_t study_session_index);
            
            // Clear the selected study session.
            void ClearStudySession();
            
            // Get the courses available for the currently selected study session.
            CourseList GetCoursesForSelectedStudySession() const;
            
            // Get a list of currently selected courses (without terms being specified).
            vector<Offerings> GetSelectedCourses() const;
            
            // Select courses (without specifying terms).
            void SelectCourses(const vector<string>& activity_keys);
            
            // Clear selected courses.
            void ClearSelectedCourses();

            // For a selected course, specify what term it is taken in.
            void SpecifyTerm(const string& activity_code, session_type term);
            
            // Generate configurations, taking into account terms specified using SpecifyTerm.
            // Returns whether the selections made using SpecifyTerm are enough to decide
            // which configurations the user wants to generate timetables for (true), or it is
            // necessary to ask the user to select among the configurations generated by
            // the configuration generation algorithm (false).
            // In the former case, the chosen configurations are auto-selected, so there is
            // no need to call SelectConfigurations().
            bool GenerateConfigurations();
            
            // Un-specify the hints that were given using SpecifyTerm and applied using
            // ApplyConfigurationHints.
            void UnspecifyTerms();
            
            // Set the set of available activity configurations to the specified ones
            // for the course load. This can be used to achieve selecting
            // multiple unbalanaced configurations.
            void SetConfigurations(const vector<ActivityConfiguration>& configs);
            
            // Get the number of configurations in the course load.
            size_t GetConfigCount() const;
            
            // Select configurations in the course load.
            void SelectConfigurations(const vector<size_t>& configs);
            
            // Get the number of selected configurations in the course load.
            size_t GetSelectedConfigCount() const;
            
            // Get the number of activities in the course load.
            size_t GetActivityCount() const;
            
            // Get information about an available configuration in the course load.
            ActivityConfiguration GetConfig(size_t config_index) const;
            
            // Get information about a selected configuration in the course load.
            ActivityConfiguration GetSelectedConfig(size_t config_index) const;
            
            // Clear selected configurations.
            void ClearSelectedConfigurations();
            
            // Generate timetables for a given configuration.
            void GenerateTimetables(ProgressIndicator progress_indicator,
                                    function<void(GenerationStatistics)> generation_stats_callback);
            
            // Get the number of timetables for a given configuration.
            size_t GetTimetableCount(size_t config_index) const;
            
            // Get a timetable.
            // The timetable lives inside the backend; it may be, doesn't have to be, copied.
            const Timetable& GetTimetable(size_t config_index, size_t timetable_index) const;
            
            // Clear the generated timetables.
            void ClearTimetables();
            
            // Get pairs of conflicting meeting sections for a given configuration.
            vector<pair<string, string>> GetConflicts(size_t config_index) const;
            
            // Get a list of criteria by which timetables for the given configuration are currently sorted.
            timetable_sort_criteria GetCurrentSortCriteria(size_t config_index) const;
            
            // Sort the timetables for a particular configuration, with the given sort criteria.
            // Use the given callback function as a progress indicator, which has the given display granularity.
            void SortTimetables(size_t config_index, const timetable_sort_criteria& sort_criteria, 
                                ProgressIndicator progress_indicator);
        private:
            // Constructor. Private because only UserSession is allowed to use it.
            CourseLoad(unique_ptr<course_load_impl>);
            friend class user_session_impl;
            
            // Instance variables
            unique_ptr<course_load_impl> impl;
        };
        
        // class UserSession
        // This is the interface to API functions using per-session state
        class UserSession
        {
        public:
            // Make class non-copyable but movable.
            // This is so that TimetableManager::StartSession() can return a UserSession by value
            // and thus avoid having to manage the lifetime of the session itself.
            UserSession(UserSession&&);
            UserSession& operator=(UserSession&&);
            
            // Destructor
            // Public because TimetableManager::StartSession() returns a UserSession by value.
            ~UserSession();
            
            // API functions
            
            // Create a new course load.
            // The course load state will be course_load_state::empty.
            CourseLoad NewCourseLoad(function<void(FileSaveStatus)> file_save_status_observer = function<void(FileSaveStatus)>());
            
            // Open a course load from a file.
            // This selects the courses specified in the course load, and sets the
            // set of available activity configurations to be the ones specified
            // in the course load, if any (as if by SetConfigurations()).
            CourseLoad OpenCourseLoad(const string& filename,
                                      function<void(FileSaveStatus)> file_save_status_observer = function<void(FileSaveStatus)>());
        private:
            // Constructor. Private because only TimetableManager is allowed to use it.
            UserSession(unique_ptr<user_session_impl>);
            friend class timetable_manager_impl;
            
            // Instance variables
            unique_ptr<user_session_impl> impl;
        };
        
        // class TimetableManager
        // This is the interface to API functions using global state and API functions for managing sessions.
        class TimetableManager
        {
        public:
            // Public interface.
            
            // Make class non-copyable
            TimetableManager(const TimetableManager&) = delete;
            
            // Creates the (only) timetable_manager instance.
            static TimetableManager& Create(const string& data_files_path = "");

            // Get a list of all study sessions.
            vector<StudySession> GetStudySessions() const;
            
            // Get the courses available for a given study session.
            CourseList GetCoursesForStudySession(size_t study_session_index) const;

            // Get a list of the filenames of the course data files.
            vector<string> GetCourseDataFileNames() const;
            
            // Update timetable data from the university websites.
            void UpdateTimetableData(ProgressIndicator progress_indicator);
            
            // Load timetable data from the specified source.
            void LoadTimetableData(TimetableDataSource& source, ProgressIndicator progress_indicator);
            
            // Reload all data used by the API. This includes inventories as well as other data
            // like study sessions. This should be called when a data file like sessions.txt or
            // altwkexcept.txt changes.
            void ReloadData();
            
            // Start a new user session.
            UserSession StartSession() const;
        private:
            // Constructor/destructor.
            // These are private because this class is a singleton.
            TimetableManager(unique_ptr<timetable_manager_impl>);
            ~TimetableManager();
            
            // Instance variables
            unique_ptr<timetable_manager_impl> impl;
        };
    
    }  // end namespace detail
    
    using detail::term_type;
    using detail::session_type;
    using detail::timetable_sort_criterion;
    using detail::timetable_sort_criteria;
    using detail::course_load_state;
    using detail::ClassTime;
    using detail::Block;
    using detail::CourseBlocks;
    using detail::TimetablePage;
    using detail::Timetable;
    using detail::Offering;
    using detail::Offerings;
    using detail::CourseList;
    using detail::ActivityConfiguration;
    using detail::StudySession;
    using detail::GenerationStatistics;
    using detail::FileSaveStatus;
    using detail::TimetableDataSource;
    using detail::TimetableDataSink;
    using detail::ProgressIndicator;
    using detail::GetEarliestClassTime;
    using detail::GetTotalTimeBlocks;
    using detail::GetTimetableSortCriteria;
    using detail::CheckUniversityWebsiteConnectivity;
    using detail::CourseLoad;
    using detail::UserSession;
    using detail::TimetableManager;
    
}  // end namespace tg

#endif  // LIBTG_INCLUDE_TG_API_HPP_INCLUDE_GUARD
